@@ -16,18 +16,33 @@
 //  You should have received a copy of the GNU General Public License
 //  along with this program; if not, write to the Free Software
 //  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-#include <Rcpp.h>
-#include <gdtools.h>
+
+extern "C" {
+#include <png.h>
+}
+
+#include <cpp11/list.hpp>
+#include <cpp11/strings.hpp>
+#include <cpp11/as.hpp>
+#include <cpp11/environment.hpp>
+#include <cpp11/external_pointer.hpp>
+#include <cpp11/protect.hpp>
 #include <string>
+#include <cstring>
 #include <iomanip>
 #include <sstream>
-#include <boost/shared_ptr.hpp>
+#include <memory>
+#include <vector>
 #include <R_ext/GraphicsEngine.h>
 
 #include "SvgStream.h"
 #include "utils.h"
+#include "tinyformat.h"
 
-typedef boost::shared_ptr<SvgStream> SvgStreamPtr;
+#include "engine_version.h"
+#include "glyph_dims.h"
+
+typedef std::shared_ptr<SvgStream> SvgStreamPtr;
 
 // SVG device metadata
 class SVGDesc {
@@ -35,19 +50,30 @@ public:
   SvgStreamPtr stream;
 
   int pageno;
+  bool is_inited;
   std::string clipid;  // ID for the clip path
   double clipx0, clipx1, clipy0, clipy1;  // Save the previous clip path to avoid duplication
   bool standalone;
-  Rcpp::List system_aliases;
-  Rcpp::List user_aliases;
+  bool always_valid;
+  const std::string file;
 
-  SVGDesc(SvgStreamPtr stream_, bool standalone_, Rcpp::List aliases_):
+  SVGDesc(SvgStreamPtr stream_, bool standalone_, const std::string& file_, bool always_valid_):
       stream(stream_),
       pageno(0),
+      is_inited(false),
       clipx0(0), clipx1(0), clipy0(0), clipy1(0),
       standalone(standalone_),
-      system_aliases(Rcpp::wrap(aliases_["system"])),
-      user_aliases(Rcpp::wrap(aliases_["user"])) {
+      always_valid(always_valid_),
+      file(file_) {
+  }
+
+  void nextFile() {
+    stream->finish(false);
+    if (stream->is_file_stream()) {
+      SvgStreamPtr newStream(new SvgStreamFile(file, pageno + 1, always_valid));
+      stream = newStream;
+    }
+    clipid.clear();
   }
 };
 
@@ -73,75 +99,116 @@ inline bool is_symbol(int face) {
   return face == 5;
 }
 
-inline std::string find_alias_field(std::string& family, Rcpp::List& alias,
-                                    const char* face, const char* field) {
-  if (alias.containsElementNamed(face)) {
-    Rcpp::List font = alias[face];
-    if (font.containsElementNamed(field))
-      return font[field];
+const static char encode_lookup[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
+const static char pad_character = '=';
+inline std::string base64_encode(const std::uint8_t* buffer, size_t size) {
+  std::string encoded_string;
+  encoded_string.reserve(((size/3) + (size % 3 > 0)) * 4);
+  std::uint32_t temp{};
+  int index = 0;
+  for (size_t idx = 0; idx < size/3; idx++) {
+    temp  = buffer[index++] << 16; //Convert to big endian
+    temp += buffer[index++] << 8;
+    temp += buffer[index++];
+    encoded_string.append(1, encode_lookup[(temp & 0x00FC0000) >> 18]);
+    encoded_string.append(1, encode_lookup[(temp & 0x0003F000) >> 12]);
+    encoded_string.append(1, encode_lookup[(temp & 0x00000FC0) >> 6 ]);
+    encoded_string.append(1, encode_lookup[(temp & 0x0000003F)      ]);
   }
-  return std::string();
-}
-
-inline std::string find_user_alias(std::string& family,
-                                   Rcpp::List const& aliases,
-                                   int face, const char* field) {
-  std::string out;
-  if (aliases.containsElementNamed(family.c_str())) {
-    Rcpp::List alias = aliases[family];
-    if (is_bolditalic(face))
-      out = find_alias_field(family, alias, "bolditalic", field);
-    else if (is_bold(face))
-      out = find_alias_field(family, alias, "bold", field);
-    else if (is_italic(face))
-      out = find_alias_field(family, alias, "italic", field);
-    else if (is_symbol(face))
-      out = find_alias_field(family, alias, "symbol", field);
-    else
-      out = find_alias_field(family, alias, "plain", field);
+  switch (size % 3) {
+  case 1:
+    temp  = buffer[index++] << 16; //Convert to big endian
+    encoded_string.append(1, encode_lookup[(temp & 0x00FC0000) >> 18]);
+    encoded_string.append(1, encode_lookup[(temp & 0x0003F000) >> 12]);
+    encoded_string.append(2, pad_character);
+    break;
+  case 2:
+    temp  = buffer[index++] << 16; //Convert to big endian
+    temp += buffer[index++] << 8;
+    encoded_string.append(1, encode_lookup[(temp & 0x00FC0000) >> 18]);
+    encoded_string.append(1, encode_lookup[(temp & 0x0003F000) >> 12]);
+    encoded_string.append(1, encode_lookup[(temp & 0x00000FC0) >> 6 ]);
+    encoded_string.append(1, pad_character);
+    break;
   }
-  return out;
+  return encoded_string;
 }
 
-inline std::string find_system_alias(std::string& family,
-                                     Rcpp::List const& aliases) {
-  std::string out;
-  if (aliases.containsElementNamed(family.c_str())) {
-    SEXP alias = aliases[family];
-    if (TYPEOF(alias) == STRSXP && Rf_length(alias) == 1)
-      out = Rcpp::as<std::string>(alias);
+static void png_memory_write(png_structp  png_ptr, png_bytep data, png_size_t length) {
+  std::vector<uint8_t> *p = (std::vector<uint8_t>*)png_get_io_ptr(png_ptr);
+  p->insert(p->end(), data, data + length);
+}
+inline std::string raster_to_string(unsigned int *raster, int w, int h, double width, double height, bool interpolate) {
+  h = h < 0 ? -h : h;
+  w = w < 0 ? -w : w;
+  bool resize = false;
+  int w_fac = 1, h_fac = 1;
+  std::vector<unsigned int> raster_resize;
+
+  if (!interpolate && double(w) < width) {
+    resize = true;
+    w_fac = std::ceil(width / w);
+  }
+  if (!interpolate && double(h) < height) {
+    resize = true;
+    h_fac = std::ceil(height / h);
   }
-  return out;
-}
 
-inline std::string fontname(const char* family_, int face,
-                            Rcpp::List const& system_aliases,
-                            Rcpp::List const& user_aliases) {
-  std::string family(family_);
-  if (face == 5)
-    family = "symbol";
-  else if (family == "")
-    family = "sans";
+  if (resize) {
+    int w_new = w * w_fac;
+    int h_new = h * h_fac;
+    raster_resize.reserve(w_new * h_new);
+    for (int i = 0; i < h; ++i) {
+      for (int j = 0; j < w; ++j) {
+        unsigned int val = raster[i * w + j];
+        for (int wrep = 0; wrep < w_fac; ++wrep) {
+          raster_resize.push_back(val);
+        }
+      }
+      for (int hrep = 1; hrep < h_fac; ++hrep) {
+        raster_resize.insert(raster_resize.end(), raster_resize.end() - w_new, raster_resize.end());
+      }
+    }
+    raster = raster_resize.data();
+    w = w_new;
+    h = h_new;
+  }
 
-  std::string alias = find_system_alias(family, system_aliases);
-  if (!alias.size())
-    alias = find_user_alias(family, user_aliases, face, "name");
+  png_structp png = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
+  if (!png) {
+    return "";
+  }
+  png_infop info = png_create_info_struct(png);
+  if (!info) {
+    png_destroy_write_struct(&png, (png_infopp)NULL);
+    return "";
+  }
+  if (setjmp(png_jmpbuf(png))) {
+    png_destroy_write_struct(&png, &info);
+    return "";
+  }
+  png_set_IHDR(
+    png,
+    info,
+    w, h,
+    8,
+    PNG_COLOR_TYPE_RGBA,
+    PNG_INTERLACE_NONE,
+    PNG_COMPRESSION_TYPE_DEFAULT,
+    PNG_FILTER_TYPE_DEFAULT
+  );
+  std::vector<uint8_t*> rows(h);
+  for (int y = 0; y < h; ++y) {
+    rows[y] = (uint8_t*)raster + y * w * 4;
+  }
 
-  if (alias.size())
-    return alias;
-  else
-    return family;
-}
+  std::vector<std::uint8_t> buffer;
+  png_set_rows(png, info, &rows[0]);
+  png_set_write_fn(png, &buffer, png_memory_write, NULL);
+  png_write_png(png, info, PNG_TRANSFORM_IDENTITY, NULL);
+  png_destroy_write_struct(&png, &info);
 
-inline std::string fontfile(const char* family_, int face,
-                            Rcpp::List user_aliases) {
-  std::string family(family_);
-  if (face == 5)
-    family = "symbol";
-  else if (family == "")
-    family = "sans";
-
-  return find_user_alias(family, user_aliases, face, "file");
+  return base64_encode(buffer.data(), buffer.size());
 }
 
 inline void write_escaped(SvgStreamPtr stream, const char* text) {
@@ -165,7 +232,7 @@ inline void write_attr_str(SvgStreamPtr stream, const char* attr, const char* va
 
 // Writing clip path attribute
 inline void write_attr_clip(SvgStreamPtr stream, std::string clipid) {
-  if (!clipid.size())
+  if (clipid.empty())
     return;
 
   (*stream) << " clip-path='url(#cp" << clipid << ")'";
@@ -203,6 +270,12 @@ inline void write_style_dbl(SvgStreamPtr stream, const char* attr, double value,
   (*stream) << attr << ": " << value << ';';
 }
 
+// Writing style attributes whose values are int type
+inline void write_style_int(SvgStreamPtr stream, const char* attr, int value, bool first = false) {
+  if(!first)  (*stream) << ' ';
+  (*stream) << attr << ": " << value << ';';
+}
+
 inline void write_style_fontsize(SvgStreamPtr stream, double value, bool first = false) {
   if(!first) (*stream) << ' ';
   // Firefox requires that we provide a unit (even though px is
@@ -225,9 +298,10 @@ inline double scale_lty(int lty, double lwd) {
 // Writing style attributes related to line types
 inline void write_style_linetype(SvgStreamPtr stream, const pGEcontext gc, bool first = false) {
   int lty = gc->lty;
+  double lwd = gc->lwd;
 
   // 1 lwd = 1/96", but units in rest of document are 1/72"
-  write_style_dbl(stream, "stroke-width", gc->lwd / 96.0 * 72, first);
+  write_style_dbl(stream, "stroke-width", lwd / 96.0 * 72, first);
 
   // Default is "stroke: #000000;" as declared in <style>
   if (!is_black(gc->col))
@@ -243,11 +317,11 @@ inline void write_style_linetype(SvgStreamPtr stream, const pGEcontext gc, bool
     // https://github.com/wch/r-source/blob/trunk/src/include/R_ext/GraphicsEngine.h#L337
     (*stream) << " stroke-dasharray: ";
     // First number
-    (*stream) << scale_lty(lty, gc->lwd);
+    (*stream) << scale_lty(lty, lwd);
     lty = lty >> 4;
     // Remaining numbers
     for(int i = 1 ; i < 8 && lty & 15; i++) {
-      (*stream) << ',' << scale_lty(lty, gc->lwd);
+      (*stream) << ',' << scale_lty(lty, lwd);
       lty = lty >> 4;
     }
     stream->put(';');
@@ -287,79 +361,98 @@ inline void write_style_linetype(SvgStreamPtr stream, const pGEcontext gc, bool
   }
 }
 
-
 // Callback functions for graphics device --------------------------------------
 
 void svg_metric_info(int c, const pGEcontext gc, double* ascent,
                      double* descent, double* width, pDevDesc dd) {
-  SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
-
-  bool is_unicode = mbcslocale;
   if (c < 0) {
-    is_unicode = true;
     c = -c;
   }
-
-  // Convert to string - negative implies unicode code point
-  char str[16];
-  if (is_unicode) {
-    Rf_ucstoutf8(str, (unsigned int) c);
+  
+  std::unordered_map<uint32_t, Dim>::const_iterator match;
+  if (gc->fontface == 5) { // Symbol
+    match = SYMBOLA_DIM.find((uint32_t) c);
+    if (match == SYMBOLA_DIM.end()) {
+      match = SYMBOLA_DIM.find(1);
+    }
   } else {
-    str[0] = (char) c;
-    str[1] = '\0';
+    match = LIBERATION_DIM.find((uint32_t) c);
+    if (match == LIBERATION_DIM.end()) {
+      match = LIBERATION_DIM.find(1);
+    }
   }
-
-  std::string font_file = fontfile(gc->fontfamily, gc->fontface, svgd->user_aliases);
-  double font_size = gc->cex * gc->ps;
-  struct fthb_string_info metrics = { 0 };
-
-  fthb_calc_string_info(str, font_file.c_str(), font_size, &metrics);
-
-  *width = metrics.width;
-  *ascent = metrics.ascent;
-  *descent = metrics.descent;
+  
+  double mod = gc->ps * gc->cex / 12;
+  
+  *ascent = match->second.ascent * mod;
+  *descent = match->second.descent * mod;
+  *width = match->second.width * mod;
 }
 
 void svg_clip(double x0, double x1, double y0, double y1, pDevDesc dd) {
   SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
+  if (!svgd->is_inited) {
+    return;
+  }
   SvgStreamPtr stream = svgd->stream;
 
+  double xmin = std::min(x0, x1);
+  double xmax = std::max(x0, x1);
+  double ymin = std::min(y0, y1);
+  double ymax = std::max(y0, y1);
+
   // Avoid duplication
-  if (std::abs(x0 - svgd->clipx0) < 0.01 &&
-      std::abs(x1 - svgd->clipx1) < 0.01 &&
-      std::abs(y0 - svgd->clipy0) < 0.01 &&
-      std::abs(y1 - svgd->clipy1) < 0.01)
+  if (std::abs(xmin - svgd->clipx0) < 0.01 &&
+      std::abs(xmax - svgd->clipx1) < 0.01 &&
+      std::abs(ymin - svgd->clipy0) < 0.01 &&
+      std::abs(ymax - svgd->clipy1) < 0.01)
     return;
 
   std::ostringstream s;
   s << std::fixed << std::setprecision(2);
-  s << dbl_format(x0) << "|" << dbl_format(x1) << "|" <<
-       dbl_format(y0) << "|" << dbl_format(y1);
-  std::string clipid = gdtools::base64_string_encode(s.str());
+  s << dbl_format(xmin) << "|" << dbl_format(xmax) << "|" <<
+    dbl_format(ymin) << "|" << dbl_format(ymax);
+  std::string str = s.str();
+  const std::uint8_t* buffer = reinterpret_cast<const std::uint8_t*>(str.data());
+  std::string clipid = base64_encode(buffer, str.size());
 
   svgd->clipid = clipid;
-  svgd->clipx0 = x0;
-  svgd->clipx1 = x1;
-  svgd->clipy0 = y0;
-  svgd->clipy1 = y1;
+  svgd->clipx0 = xmin;
+  svgd->clipx1 = xmax;
+  svgd->clipy0 = ymin;
+  svgd->clipy1 = ymax;
+
+  if (stream->is_clipping()) {
+    (*stream) << "</g>\n";
+  }
+  // Is this clip region already defined?
+  if (!stream->has_clip_id(clipid)) {
+    stream->add_clip_id(clipid);
+
+    (*stream) << "<defs>\n";
+    (*stream) << "  <clipPath id='cp" << svgd->clipid << "'>\n";
+    (*stream) << "    <rect x='" << xmin << "' y='" << ymin <<
+      "' width='" << (xmax - xmin) << "' height='" << (ymax - ymin) << "' />\n";
+    (*stream) << "  </clipPath>\n";
+    (*stream) << "</defs>\n";
+  }
+
+  (*stream) << "<g";
+  write_attr_clip(stream, svgd->clipid);
+  (*stream) << ">\n";
 
-  (*stream) << "<defs>\n";
-  (*stream) << "  <clipPath id='cp" << svgd->clipid << "'>\n";
-  (*stream) << "    <rect x='" << std::min(x0, x1) << "' y='" << std::min(y0, y1) <<
-    "' width='" << std::abs(x1 - x0) << "' height='" << std::abs(y1 - y0) << "' />\n";
-  (*stream) << "  </clipPath>\n";
-  (*stream) << "</defs>\n";
   stream->flush();
 }
 
 void svg_new_page(const pGEcontext gc, pDevDesc dd) {
-BEGIN_RCPP
-
   SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
   SvgStreamPtr stream = svgd->stream;
 
   if (svgd->pageno > 0) {
-    Rcpp::stop("svglite only supports one page");
+
+    // close existing file, create a new one, and update stream
+    svgd->nextFile();
+    stream = svgd->stream;
   }
 
   if (svgd->standalone)
@@ -372,18 +465,18 @@ BEGIN_RCPP
     (*stream) << " xmlns:xlink='http://www.w3.org/1999/xlink'";
   }
 
+  (*stream) << " class='svglite'";
+  
+  (*stream) << " data-engine-version='" << get_engine_version() << "'";
+
+  (*stream) << " width='" << dd->right << "pt' height='" << dd->bottom << "pt'";
+
   (*stream) << " viewBox='0 0 " << dd->right << ' ' << dd->bottom << "'>\n";
-
-  // Initialise clipping the same way R does
-  svgd->clipx0 = 0;
-  svgd->clipy0 = dd->bottom;
-  svgd->clipx1 = dd->right;
-  svgd->clipy1 = 0;
-
+  
   // Setting default styles
   (*stream) << "<defs>\n";
   (*stream) << "  <style type='text/css'><![CDATA[\n";
-  (*stream) << "    line, polyline, polygon, path, rect, circle {\n";
+  (*stream) << "    .svglite line, .svglite polyline, .svglite polygon, .svglite path, .svglite rect, .svglite circle {\n";
   (*stream) << "      fill: none;\n";
   (*stream) << "      stroke: #000000;\n";
   (*stream) << "      stroke-linecap: round;\n";
@@ -403,21 +496,35 @@ BEGIN_RCPP
   write_style_end(stream);
   (*stream) << "/>\n";
 
+
+  // Initialise clipping - make sure the stored clipping is bogus to force
+  // svg_clip to do its thing
+  svgd->clipx0 = R_PosInf;
+  svgd->clipy0 = R_NegInf;
+  svgd->clipx1 = R_NegInf;
+  svgd->clipy1 = R_PosInf;
+  svgd->is_inited = true;
+
+  svg_clip(0, dd->right, dd->bottom, 0, dd);
+
   svgd->stream->flush();
   svgd->pageno++;
-
-VOID_END_RCPP
 }
 
 void svg_close(pDevDesc dd) {
   SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
-  svgd->stream->finish();
+  if (svgd->is_inited) {
+    svgd->stream->finish(true);
+  }
   delete(svgd);
 }
 
 void svg_line(double x1, double y1, double x2, double y2,
               const pGEcontext gc, pDevDesc dd) {
   SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
+  if (!svgd->is_inited) {
+    return;
+  }
   SvgStreamPtr stream = svgd->stream;
 
   (*stream) << "<line x1='" << x1 << "' y1='" << y1 << "' x2='" <<
@@ -427,8 +534,6 @@ void svg_line(double x1, double y1, double x2, double y2,
   write_style_linetype(stream, gc, true);
   write_style_end(stream);
 
-  write_attr_clip(stream, svgd->clipid);
-
   (*stream) << " />\n";
   stream->flush();
 }
@@ -437,6 +542,9 @@ void svg_poly(int n, double *x, double *y, int filled, const pGEcontext gc,
               pDevDesc dd, const char* node_name) {
 
   SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
+  if (!svgd->is_inited) {
+    return;
+  }
   SvgStreamPtr stream = svgd->stream;
 
   (*stream) << "<" << node_name << " points='";
@@ -452,8 +560,6 @@ void svg_poly(int n, double *x, double *y, int filled, const pGEcontext gc,
     write_style_col(stream, "fill", gc->fill);
   write_style_end(stream);
 
-  write_attr_clip(stream, svgd->clipid);
-
   (*stream) << " />\n";
   stream->flush();
 
@@ -473,6 +579,9 @@ void svg_path(double *x, double *y,
               Rboolean winding,
               const pGEcontext gc, pDevDesc dd) {
   SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
+  if (!svgd->is_inited) {
+    return;
+  }
   SvgStreamPtr stream = svgd->stream;
 
   // Create path data
@@ -501,27 +610,45 @@ void svg_path(double *x, double *y,
   write_style_linetype(stream, gc);
   write_style_end(stream);
 
-  write_attr_clip(stream, svgd->clipid);
-
   (*stream) << " />\n";
   stream->flush();
 }
 
 double svg_strwidth(const char *str, const pGEcontext gc, pDevDesc dd) {
-  SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
-
-  std::string font_file = fontfile(gc->fontfamily, gc->fontface, svgd->user_aliases);
-  double font_size = gc->cex * gc->ps;
-  double string_width = 0.0;
-
-  fthb_calc_string_width(str, font_file.c_str(), font_size, &string_width);
-
-  return string_width;
+  static UTF_UCS converter;
+  
+  double width = 0.0;
+  
+  int n_chars = 0;
+  
+  uint32_t* chars = converter.convert(str, n_chars);
+  
+  std::unordered_map<uint32_t, Dim>::const_iterator match;
+  for (int i = 0; i < n_chars; ++i) {
+    if (gc->fontface == 5) { // Symbol
+      match = SYMBOLA_DIM.find(chars[i]);
+      if (match == SYMBOLA_DIM.end()) {
+        match = SYMBOLA_DIM.find(1);
+      }
+    } else {
+      match = LIBERATION_DIM.find(chars[i]);
+      if (match == LIBERATION_DIM.end()) {
+        match = LIBERATION_DIM.find(1);
+      }
+    }
+    width += match->second.width;
+  }
+  double mod = gc->ps * gc->cex / 12;
+  
+  return width * mod;
 }
 
 void svg_rect(double x0, double y0, double x1, double y1,
               const pGEcontext gc, pDevDesc dd) {
   SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
+  if (!svgd->is_inited) {
+    return;
+  }
   SvgStreamPtr stream = svgd->stream;
 
   // x and y give top-left position
@@ -534,8 +661,6 @@ void svg_rect(double x0, double y0, double x1, double y1,
     write_style_col(stream, "fill", gc->fill);
   write_style_end(stream);
 
-  write_attr_clip(stream, svgd->clipid);
-
   (*stream) << " />\n";
   stream->flush();
 }
@@ -543,9 +668,12 @@ void svg_rect(double x0, double y0, double x1, double y1,
 void svg_circle(double x, double y, double r, const pGEcontext gc,
                        pDevDesc dd) {
   SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
+  if (!svgd->is_inited) {
+    return;
+  }
   SvgStreamPtr stream = svgd->stream;
 
-  (*stream) << "<circle cx='" << x << "' cy='" << y << "' r='" << r << "pt'";
+  (*stream) << "<circle cx='" << x << "' cy='" << y << "' r='" << r << "'";
 
   write_style_begin(stream);
   write_style_linetype(stream, gc, true);
@@ -553,8 +681,6 @@ void svg_circle(double x, double y, double r, const pGEcontext gc,
     write_style_col(stream, "fill", gc->fill);
   write_style_end(stream);
 
-  write_attr_clip(stream, svgd->clipid);
-
   (*stream) << " />\n";
   stream->flush();
 }
@@ -562,15 +688,10 @@ void svg_circle(double x, double y, double r, const pGEcontext gc,
 void svg_text(double x, double y, const char *str, double rot,
               double hadj, const pGEcontext gc, pDevDesc dd) {
   SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
-  SvgStreamPtr stream = svgd->stream;
-
-  // If we specify the clip path inside <text>, the "transform" also
-  // affects the clip path, so we need to specify clip path at an outer level
-  if (svgd->clipid.size()) {
-    (*stream) << "<g";
-    write_attr_clip(stream, svgd->clipid);
-    stream->put('>');
+  if (!svgd->is_inited) {
+    return;
   }
+  SvgStreamPtr stream = svgd->stream;
 
   (*stream) << "<text";
 
@@ -584,35 +705,37 @@ void svg_text(double x, double y, const char *str, double rot,
 
   double fontsize = gc->cex * gc->ps;
 
+  if (hadj == 0.5) {
+    write_attr_str(stream, "text-anchor", "middle");
+  } else if (hadj == 1) {
+    write_attr_str(stream, "text-anchor", "end");
+  }
+
   write_style_begin(stream);
   write_style_fontsize(stream, fontsize, true);
-  if (is_bold(gc->fontface))
+
+  if (is_bold(gc->fontface)) {
     write_style_str(stream, "font-weight", "bold");
+  }
   if (is_italic(gc->fontface))
     write_style_str(stream, "font-style", "italic");
   if (!is_black(gc->col))
     write_style_col(stream, "fill", gc->col);
+  
+  write_style_str(stream, "font-family", "sans");
 
-  std::string font_name = fontname(gc->fontfamily, gc->fontface, svgd->system_aliases, svgd->user_aliases);
-  write_style_str(stream, "font-family", font_name.c_str());
   write_style_end(stream);
 
-  double string_width = 0.0;
-
-  std::string font_file = fontfile(gc->fontfamily, gc->fontface, svgd->user_aliases);
-  fthb_calc_string_width(str, font_file.c_str(), fontsize, &string_width);
-
-  (*stream) << " textLength='" << string_width << "px'";
+  double width = svg_strwidth(str, gc, dd);
+  (*stream) << " textLength='" << width << "px'";
   (*stream) << " lengthAdjust='spacingAndGlyphs'";
+  
   stream->put('>');
 
   write_escaped(stream, str);
 
   (*stream) << "</text>";
 
-  if (svgd->clipid.size())
-    (*stream) << "</g>";
-
   stream->put('\n');
   stream->flush();
 }
@@ -632,32 +755,25 @@ void svg_raster(unsigned int *raster, int w, int h,
                 Rboolean interpolate,
                 const pGEcontext gc, pDevDesc dd) {
   SVGDesc *svgd = (SVGDesc*) dd->deviceSpecific;
+  if (!svgd->is_inited) {
+    return;
+  }
   SvgStreamPtr stream = svgd->stream;
 
   if (height < 0)
     height = -height;
 
-  std::vector<unsigned int> raster_(w*h);
-  for (std::vector<unsigned int>::size_type i = 0 ; i < raster_.size(); ++i) {
-    raster_[i] = raster[i] ;
-  }
-
-  std::string base64_str = gdtools::raster_to_str(raster_, w, h, width, height,
-    (Rboolean) interpolate);
-
-  // If we specify the clip path inside <image>, the "transform" also
-  // affects the clip path, so we need to specify clip path at an outer level
-  if (svgd->clipid.size()) {
-    (*stream) << "<g";
-    write_attr_clip(stream, svgd->clipid);
-    stream->put('>');
-  }
+  std::string base64_str = raster_to_string(raster, w, h, width, height, interpolate);
 
   (*stream) << "<image";
   write_attr_dbl(stream, "width", width);
   write_attr_dbl(stream, "height", height);
   write_attr_dbl(stream, "x", x);
   write_attr_dbl(stream, "y", y - height);
+  write_attr_str(stream, "preserveAspectRatio", "none");
+  if (!interpolate) {
+    write_attr_str(stream, "image-rendering", "pixelated");
+  }
 
   if( rot != 0 ){
     (*stream) << tfm::format(" transform='rotate(%0.0f,%.2f,%.2f)'", -1.0 * rot, x, y);
@@ -666,9 +782,6 @@ void svg_raster(unsigned int *raster, int w, int h,
   (*stream) << " xlink:href='data:image/png;base64," << base64_str << '\'';
   (*stream) << "/>";
 
-  if (svgd->clipid.size())
-    (*stream) << "</g>";
-
   stream->put('\n');
   stream->flush();
 }
@@ -677,7 +790,7 @@ SEXP svg_set_pattern(SEXP pattern, pDevDesc dd) {
     return R_NilValue;
 }
 
-void svg_release_pattern(SEXP ref, pDevDesc dd) {} 
+void svg_release_pattern(SEXP ref, pDevDesc dd) {}
 
 SEXP svg_set_clip_path(SEXP path, SEXP ref, pDevDesc dd) {
     return R_NilValue;
@@ -693,7 +806,8 @@ void svg_release_mask(SEXP ref, pDevDesc dd) {}
 
 pDevDesc svg_driver_new(SvgStreamPtr stream, int bg, double width,
                         double height, double pointsize,
-                        bool standalone, Rcpp::List& aliases) {
+                        bool standalone, const std::string& file, 
+                        bool always_valid) {
 
   pDevDesc dd = (DevDesc*) calloc(1, sizeof(DevDesc));
   if (dd == NULL)
@@ -733,7 +847,7 @@ pDevDesc svg_driver_new(SvgStreamPtr stream, int bg, double width,
   dd->setMask         = svg_set_mask;
   dd->releaseMask     = svg_release_mask;
 #endif
- 
+
   // UTF-8 support
   dd->wantSymbolUTF8 = (Rboolean) 1;
   dd->hasTextUTF8 = (Rboolean) 1;
@@ -760,7 +874,7 @@ pDevDesc svg_driver_new(SvgStreamPtr stream, int bg, double width,
 
   // Capabilities
   dd->canClip = TRUE;
-  dd->canHAdj = 0;
+  dd->canHAdj = 1;
   dd->canChangeGamma = FALSE;
   dd->displayListOn = FALSE;
   dd->haveTransparency = 2;
@@ -770,12 +884,13 @@ pDevDesc svg_driver_new(SvgStreamPtr stream, int bg, double width,
   dd->deviceVersion = R_GE_definitions;
 #endif
 
-  dd->deviceSpecific = new SVGDesc(stream, standalone, aliases);
+  dd->deviceSpecific = new SVGDesc(stream, standalone, file, always_valid);
   return dd;
 }
 
 void makeDevice(SvgStreamPtr stream, std::string bg_, double width, double height,
-                double pointsize, bool standalone, Rcpp::List& aliases) {
+                double pointsize, bool standalone, const std::string& file,
+                bool always_valid) {
 
   int bg = R_GE_str2col(bg_.c_str());
 
@@ -783,9 +898,9 @@ void makeDevice(SvgStreamPtr stream, std::string bg_, double width, double heigh
   R_CheckDeviceAvailable();
   BEGIN_SUSPEND_INTERRUPTS {
     pDevDesc dev = svg_driver_new(stream, bg, width, height, pointsize,
-                                  standalone, aliases);
+                                  standalone, file, always_valid);
     if (dev == NULL)
-      Rcpp::stop("Failed to start SVG device");
+      cpp11::stop("Failed to start SVG device");
 
     pGEDevDesc dd = GEcreateDevDesc(dev);
     GEaddDevice2(dd, "devSVG");
@@ -794,37 +909,38 @@ void makeDevice(SvgStreamPtr stream, std::string bg_, double width, double heigh
   } END_SUSPEND_INTERRUPTS;
 }
 
-// [[Rcpp::export]]
+[[cpp11::register]]
 bool svglite_(std::string file, std::string bg, double width, double height,
-              double pointsize, bool standalone, Rcpp::List aliases) {
+              double pointsize, bool standalone, bool always_valid) {
 
-  SvgStreamPtr stream(new SvgStreamFile(file));
-  makeDevice(stream, bg, width, height, pointsize, standalone, aliases);
+  SvgStreamPtr stream(new SvgStreamFile(file, 1, always_valid));
+  makeDevice(stream, bg, width, height, pointsize, standalone, file, always_valid);
 
   return true;
 }
 
-// [[Rcpp::export]]
-Rcpp::XPtr<std::stringstream> svgstring_(Rcpp::Environment env, std::string bg,
+[[cpp11::register]]
+cpp11::external_pointer<std::stringstream> svgstring_(cpp11::environment env, std::string bg,
                                          double width, double height, double pointsize,
-                                         bool standalone, Rcpp::List aliases) {
+                                         bool standalone) {
 
   SvgStreamPtr stream(new SvgStreamString(env));
-  makeDevice(stream, bg, width, height, pointsize, standalone, aliases);
+  makeDevice(stream, bg, width, height, pointsize, standalone, "", true);
 
   SvgStreamString* strstream = static_cast<SvgStreamString*>(stream.get());
 
-  return strstream->string_src();
+  return {strstream->string_src(), false};
 }
 
-// [[Rcpp::export]]
-std::string get_svg_content(Rcpp::XPtr<std::stringstream> p) {
+[[cpp11::register]]
+std::string get_svg_content(cpp11::external_pointer<std::stringstream> p) {
   p->flush();
   std::string svgstr = p->str();
   // If the current SVG is empty, we also make the string empty
   // Otherwise append "</svg>" to make it a valid SVG
   if(!svgstr.empty()) {
-    svgstr.append("</svg>");
+    svgstr.append("</g>\n</svg>");
   }
   return svgstr;
 }
+
